@@ -37,6 +37,8 @@ void pipe_info_close_all(pipe_info_t const *const ipipe,
 void pipe_info_dup_in_pipes(pipe_info_t *ipipe, unsigned long pipe_cnt,
                             char *cmd_name, int close_unused);
 void pipe_info_print(pipe_info_t const *const ipipe, unsigned long const cnt);
+void pipe_info_check_for_duplicates(
+        pipe_info_t const *const ipipe, unsigned long const cnt);
 
 unsigned int pipe_info_clp_count(int const start_argc, int const argc,
                                  char *const argv[]);
